#!/usr/bin/python
from __future__ import print_function
import mysql.connector
from mysql.connector import errorcode
import json
from decimal import Decimal
from boConfig import DB_USER, DB_PASS, DB_HOST, DB_PORT, DB_NAME, LOGLEVEL
import logging
import sys
from geographiclib.geodesic import Geodesic

TABLES = {}
TABLES['signals'] = (
    "CREATE TABLE `signals` ("
    "  `id` int(11) UNSIGNED NOT NULL AUTO_INCREMENT,"
    "  `time_ns` bigint(11) UNSIGNED NOT NULL,"
    "  `channel0` tinyint(1) DEFAULT NULL,"
    "  `channel1` tinyint(1) DEFAULT NULL,"
    "  `channel2` tinyint(1) DEFAULT NULL,"
    "  `channel3` tinyint(1) DEFAULT NULL,"
    "  PRIMARY KEY (`id`),"
    "  UNIQUE KEY `id` (`id`),"
    "  UNIQUE KEY `time_ns` (`time_ns`)"
    ") ENGINE=InnoDB DEFAULT CHARSET=utf8;"
    )

TABLES['strokes'] = (
    "CREATE TABLE `strokes` ("
    "  `id` int(11) UNSIGNED NOT NULL AUTO_INCREMENT,"
    "  `signal_id` INT(11) NULL DEFAULT NULL,"
    "  `time_ns_stroke` bigint(11) UNSIGNED NOT NULL,"
    "  `lat_stroke` decimal(9,6) NOT NULL,"
    "  `lon_stroke` decimal(9,6) NOT NULL,"
    "  `delta_time_ns_station` INT(11) NOT NULL,"
    "  `lat_station` decimal(9,6) NOT NULL,"
    "  `lon_station` decimal(9,6) NOT NULL,"
    "  `distance` decimal(6,1) NOT NULL,"
    "  `azimuth` DECIMAL(4,1) NOT NULL,"
    "  PRIMARY KEY (`id`),"
    "  UNIQUE KEY `id` (`id`),"
    "  UNIQUE KEY `signal_id` (`signal_id`),"
    "  UNIQUE KEY `time_ns_stroke_delta_station` (`time_ns_stroke`,`delta_time_ns_station`)"
    ") ENGINE=InnoDB DEFAULT CHARSET=utf8;"
    )

cnx = None
cursor = None

def connectMysql():
    global cnx
    global cursor
    cnx = mysql.connector.connect(user = DB_USER, password = DB_PASS, host = DB_HOST, port = DB_PORT)
    cursor = cnx.cursor()

def disconnectMysql():
    global cnx
    global cursor
    if (None != cursor):
        cursor.close()
        cursor = None
    if (None != cnx):
        cnx.commit()
        cnx.close()
        cnx = None

def createDatabase():
    global cursor
    try:
        logger.info("Creating database {}: ".format(DB_NAME))
        cursor.execute("CREATE DATABASE {} DEFAULT CHARACTER SET 'utf8'".format(DB_NAME))
    except mysql.connector.Error as err:
        logger.critical("Failed creating database {}: '{}'".format(DB_NAME, err))
        exit(1)

def createTables():
    global cursor
    for name, ddl in TABLES.iteritems():
        try:
            logger.info("Creating table {}: ".format(name))
            cursor.execute(ddl)
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_TABLE_EXISTS_ERROR:
                logger.info("Table {} already exists.".format(name))
            else:
                logger.critical("Failed creating table {}: '{}'".format(name, err))
                exit(1)

def updateDatabase_0_2():
    def getMissingAzimuthEntries():
        return queryTable("`id`, `lat_stroke`, `lon_stroke`, `lat_station`, `lon_station`", "`strokes`", "`azimuth` = 0.0 LIMIT 1000")

    def calcDistanceAndAzimuth(stroke):
        lat1 = float(stroke["lat_station"])
        lon1 = float(stroke["lon_station"])
        lat2 = float(stroke["lat_stroke"])
        lon2 = float(stroke["lon_stroke"])
        geodesic = Geodesic.WGS84.Inverse(lat1, lon1, lat2, lon2)
        distance = geodesic["s12"] # distance between station and stroke (meters)
        distance = round((distance / 1000), 1) # distance between station and stroke (km)
        azimuth = round(geodesic["azi1"], 1) # azimuth at station pointing to the stroke (degrees) in the range [-180, 180]
        if (azimuth <= 0): # exclude 0.0 as 0.0 is the marker for not yet calculated
            azimuth += 360 # azimuth at station pointing to the stroke (degrees) in the range ]0, 360]
        return distance, azimuth

    def updateDistanceAndAzimuth(stroke):
        global cursor
        done = False
        sql = "UPDATE `strokes` SET `distance` = {}, `azimuth` = {} WHERE id = {}".format(stroke["distance"], stroke["azimuth"], stroke["id"])
        try:
            cursor.execute(sql)
            done = True
        except mysql.connector.Error as err:
            logger.warning(err)
        return done

    global cursor
    logger.info("Updating database {}: ".format(DB_NAME))

    try:
        sql = "ALTER TABLE `strokes` DROP `alt_stroke`, DROP `alt_station`"
        logger.info(sql)
        cursor.execute(sql)
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_CANT_DROP_FIELD_OR_KEY:
            logger.info("`alt_stroke` or `alt_station` is alredy dropped")
        else:
            logger.critical("Failed '{}': '{}'".format(sql, err))
            exit(1)

    try:
        sql = "ALTER TABLE `strokes` ADD `azimuth` DECIMAL(4,1) NOT NULL AFTER `distance`"
        logger.info(sql)
        cursor.execute(sql)
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_DUP_FIELDNAME:
            logger.info("`azimuth` is alredy inserted")
        else:
            logger.critical("Failed '{}': '{}'".format(sql, err))
            exit(1)

    strokes = getMissingAzimuthEntries()
    updated = 0
    failed = 0
    while strokes:
        for stroke in strokes:
            stroke["distance"], stroke["azimuth"] = calcDistanceAndAzimuth(stroke)
            if updateDistanceAndAzimuth(stroke):
                updated += 1
            else:
                failed += 1
                print("failed to update {}".format(stroke))
        cnx.commit()
        print("updated = {}, failed = {}".format(updated, failed))
        strokes = getMissingAzimuthEntries()
    logger.info("updated = {}, failed = {}".format(updated, failed))

def updateDatabase():
    global DB_NAME
    logger.info("-> updateDatabase()")
    connectMysql()
    connectDatabase()
    updateDatabase_0_2()
    disconnectMysql()
    logger.info("<- updateDatabase()")

def connectDatabase():
    global cnx
    try:
        cnx.database = DB_NAME
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_BAD_DB_ERROR:
            createDatabase()
            cnx.database = DB_NAME
            createTables()
        else:
            logger.critical("Failed to connect to database {}: '{}'".format(DB_NAME, err))
            exit(1)

def insertSignal(signal):
    global cursor
    done = False
    if ("time_ns" in signal) and \
       ("channel" in signal) and \
       (4 == len(signal["channel"])) and \
       ( \
	   (({} != signal["channel"][0]) and ("data" in signal["channel"][0])) or \
           (({} != signal["channel"][1]) and ("data" in signal["channel"][1])) or \
           (({} != signal["channel"][2]) and ("data" in signal["channel"][2])) or \
           (({} != signal["channel"][3]) and ("data" in signal["channel"][3])) \
        ):
        insert = "`time_ns`"
        values = "{}".format(signal["time_ns"])
        for i in range(len(signal["channel"])):
            channel = signal["channel"][i]
            if ({} != channel):
                insert += ", `channel{}`".format(i)
                values += ", 1"
        sql = "INSERT INTO `signals` ({}) VALUES({})".format(insert, values)
        try:
            cursor.execute(sql)
            cnx.commit()
            done = True
        except mysql.connector.Error as err:
            if "Duplicate entry" == err.msg[0:15]:
                done = True
            else:
                logger.warning(err)
    else:
        logger.error("insertSignal(): at least 'time_ns' and one 'channel'[0|1|2|3] must contain 'data'")
        logger.error(signal)
    return done

def insertStroke(stroke):
    global cursor
    done = False
    keys = ["time_ns_stroke", "lat_stroke", "lon_stroke", "delta_time_ns_station", "lat_station", "lon_station", "distance", "azimuth"]
    insert = ""
    values = ""
    for key in keys:
        if (key in stroke):
            insert += "`{}`, ".format(key)
            values += "{}, ".format(stroke[key])
        else:
            logger.error("key '{}' not found in stroke".format(key))
            return done
    insert = insert[:-2] # remove trailing ", "
    values = values[:-2] # remove trailing ", "
    sql = "INSERT INTO `strokes` ({}) VALUES({})".format(insert, values)
    try:
        cursor.execute(sql)
        cnx.commit()
        done = True
    except mysql.connector.Error as err:
        if "Duplicate entry" == err.msg[0:15]:
            done = True
        else:
            logger.warning(err)
    return done

def updateStrokeSignal(link):
    global cursor
    done = False
    sql = "UPDATE `strokes` SET signal_id = {} WHERE id = {}".format(link["signal_id"], link["stroke_id"])
    try:
        cursor.execute(sql)
        cnx.commit()
        done = True
    except mysql.connector.Error as err:
        logger.warning(err)
    return done

def queryTable(select, table, where):
    global cursor
    result = []
    sql = "SELECT {} FROM {} WHERE {}".format(select, table, where)
    try:
        cursor.execute(sql)
        columns = cursor.description 
        result = [{columns[index][0]:column for index, column in enumerate(value)} for value in cursor.fetchall()]
        for i in range(len(result)):
            for key in result[i]:
                if (type(Decimal(0)) == type(result[i][key])):
                    result[i][key] = float(result[i][key])
    except mysql.connector.Error as err:
        logger.warning(err)
    return result

def querySignals(select = "*", where = 1):
    return queryTable(select, "`signals`", where)

def queryStrokes(select = "*", where = 1):
    return queryTable(select, "`strokes`", where)

def correlateSignalsStrokes(timeCorrection = -1000000000): # by manual correlation I found the data on the server being systematically 1 sec = 1000000000 ns off; in one case it was 1 ns less: 999999999 ns
    select = "signals.id as signal_id, strokes.id as stroke_id"
    tables = "`signals`, `strokes`"
    where = "strokes.time_ns_stroke + strokes.delta_time_ns_station + {} = signals.time_ns AND strokes.signal_id IS NULL".format(timeCorrection)
    return queryTable(select, tables, where)

def testInsertSignal():
    logger.info("-> testInsertSignal()")
    logger.info("# test error: only time_ns present")
    signal = {'time_ns': 12345678901}
    insertSignal(signal)

    logger.info("# test error: only channel present")
    signal = {'channel': [{}, {}, {}, {}]}
    insertSignal(signal)

    logger.info("# test error: both present but all channels empty")
    signal = {'time_ns': 12345678902, 'channel': [{}, {}, {}, {}]}
    insertSignal(signal)

    logger.info("# test error: both present but no data")
    signal = {'time_ns': 12345678903, 'channel': [{'start': 1}, {}, {}, {}]}
    insertSignal(signal)

    logger.info("# test valid: 1 set of data")
    signal = {'time_ns': 12345678904, 'channel': [{'data': "0x7F ..."}, {}, {}, {}]}
    insertSignal(signal)

    logger.info("# test valid: 1 set of data")
    signal = {'time_ns': 12345678905, 'channel': [{}, {'data': "0x7F ..."}, {}, {}]}
    insertSignal(signal)

    logger.info("# test valid: 4 sets of data")
    signal = {'time_ns': 12345678906, 'channel': [{'data': "0x7F ..."}, {'data': "0x7F ..."}, {'data': "0x7F ..."}, {'data': "0x7F ..."}]}
    insertSignal(signal)

    logger.info("# test silently ignore repeated key 'time_ns'")
    signal = {'time_ns': 12345678906, 'channel': [{'data': "0x7F ..."}, {'data': "0x7F ..."}, {'data': "0x7F ..."}, {'data': "0x7F ..."}]}
    insertSignal(signal)
    logger.info("<- testInsertSignal()")

def testQuerySignals():
    logger.info("-> testQuerySignals()")
    logger.info("# test default")
    result = querySignals()
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause")
    result = querySignals(select = "`id`, `time_ns`", where = "`channel0` = 1")
    logger.debug(json.dumps(result, indent=4))
    logger.info("<- testQuerySignals()")

def testInsertStroke():
    logger.info("-> testInsertStroke()")

    logger.info("# test error: key 'time_ns_stroke' missing")
    stroke = {'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'lat_stroke' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'lon_stroke' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'delta_time_ns_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'lat_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'lon_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'distance' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test error: key 'azimuth' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2}
    insertStroke(stroke)

    logger.info("# test valid 1")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test valid 2")
    stroke = {'time_ns_stroke': 1483871710203059999L, 'lat_stroke': 37.980776, 'lon_stroke': 24.618727, 'delta_time_ns_station': 7133193, 'lat_station': 50.348630, 'lon_station': 4.24393, 'distance': 2119.3, 'azimuth': 0.0}
    insertStroke(stroke)

    logger.info("# test silently ignore repeated key 'time_ns_stroke' + 'delta_time_ns_station'")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    insertStroke(stroke)
    logger.info("<- testInsertStroke()")

def testQueryStrokes():
    logger.info("-> testQueryStrokes()")
    logger.info("# test default")
    result = queryStrokes()
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause, no data")
    result = queryStrokes(select = "`id`, `distance`, `azimuth`", where = "`time_ns_stroke` = 0")
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause, data")
    result = queryStrokes(select = "`id`, `distance`, `azimuth`", where = "`time_ns_stroke` = 1483871710203059998")
    logger.debug(json.dumps(result, indent=4))
    logger.info("<- testQueryStrokes()")

def test():
    global DB_NAME
    DB_NAME = 'testBoSignals'
    logger.info("-> test()")

    # test connect
    connectMysql()
    connectDatabase()

    # test updating the databse
    updateDatabase_0_2()

    # tests for insert and query
    testInsertSignal()
    testQuerySignals()
    testInsertStroke()
    testQueryStrokes()

    # test disconnect
    disconnectMysql()
    logger.info("<- test()")

def configLogger():
    global logger
    if __name__ == "__main__":
        format = "%(levelname)s: %(message)s"
        logging.basicConfig(format = format)
    logger = logging.getLogger(__name__)
    if __name__ == "__main__":
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(LOGLEVEL)

def main():
    configLogger()
    if (2 == len(sys.argv)) and ("--updateDatabase" == sys.argv[1]):
        updateDatabase()
    else:
        test() # optionally enable test in order to run some quick tests on a pure test database

if __name__ == "__main__":
    main()
